/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "network/network.pmh"

namespace fe
{

ChunkReceiver::ChunkReceiver(void)
{
	m_operating = false;
	m_live = false;
}

ChunkReceiver::~ChunkReceiver(void)
{
fprintf(stderr, "pre stop\n");
	stop();
fprintf(stderr, "post stop\n");
}

void ChunkReceiver::start(sp<ChunkSpool> spChunkSpool, sp<Socket> spSocket)
{
	m_chunkReceiverTask.m_spChunkSpool = spChunkSpool;
	m_chunkReceiverTask.m_spSocket = spSocket;
	spSocket->setPoison(&m_poison);
	m_chunkReceiverTask.m_pChunkReceiver = this;
	m_live = true;
	m_pChunkReceiverThread = new Thread(&m_chunkReceiverTask);
	m_operating = true;
}

void ChunkReceiver::stop(void)
{
	if(m_operating)
	{
		m_poison.start();
fprintf(stderr, "pre join\n");
		m_pChunkReceiverThread->join();
fprintf(stderr, "post join\n");
		delete m_pChunkReceiverThread;
		m_operating = false;
		m_poison.stop();
	}
}

bool ChunkReceiver::live(void)
{
	Mutex::Guard guard(m_liveMutex);
	return m_live;
}

void ChunkReceiver::setLive(bool setting)
{
	Mutex::Guard guard(m_liveMutex);
	m_live = setting;
}

void ChunkReceiver::setPolling(sp<SignalerI> a_spSignaler,
	sp<Layout> a_spPollLayout)
{
	m_spSignaler=a_spSignaler;
	m_poll=(m_spSignaler.isValid() && a_spPollLayout.isValid())?
			a_spPollLayout->scope()->createRecord(a_spPollLayout):
			Record();
}

void ChunkReceiver::poll(void)
{
	if(m_spSignaler.isValid())
	{
		feLog("ChunkReceiver::poll\n");
		m_spSignaler->signal(m_poll);
	}
}

ChunkReceiverTask::ChunkReceiverTask(void)
{
	m_pChunkReceiver = NULL;
}

void ChunkReceiverTask::operate(void)
{
	try
	{
		m_pChunkReceiver->setLive(true);
		sp<Chunk> spChunk;
		while(true)
		{
			IWORD n;
			U32 chunkLength;
			n = m_spSocket->read((void *)&chunkLength, sizeof(U32));
			if(n != sizeof(U32))
			{
				feX(e_readFailed, "ChunkReceiverTask",
						"connection failed in size read %u vs %u",
						n,sizeof(U32));
			}
			chunkLength = ntohl(chunkLength);
			spChunk = new Chunk(chunkLength);
			feLogGroup("network","chunk length to read %u %u\n",
					chunkLength, spChunk->size());
			n = m_spSocket->read(spChunk->buffer(), spChunk->size());
			feLogGroup("network","size read %u\n", n);
			if(n != (IWORD)chunkLength)
			{
				feX(e_readFailed, "ChunkReceiverTask",
					"connection failed in chunk read");
			}
			m_spChunkSpool->put(spChunk.abandon());
			spChunk = NULL;
			m_pChunkReceiver->poll();
		}
	}
	catch(Poisoned &p)
	{
		m_pChunkReceiver->setLive(false);
		feLogGroup("network","ChunkReceiverTask poisoned\n");
	}
	catch(Exception &e)
	{
		m_pChunkReceiver->setLive(false);
		if(e.getResult() == e_readFailed)
		{
			feLogGroup("network","thread dying due to read failure\n");
			feLogGroup("network","  %s\n", e.getMessage().c_str());
		}
		else
		{
			feLogGroup("network","thread dying due to unknown failure\n");
#if FE_CODEGEN<=FE_DEBUG
			exit(1);
#else
			throw e;
#endif
		}
	}
}

} /* namespace */

