/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "xml/xml.pmh"

#define FE_SAX_DEBUG		FALSE

namespace fe
{

BWORD SurfaceAccessibleXml::load(String a_filename,Real a_frame,
	sp<StringFilterI> a_spFilter)
{
#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load \"%s\"\n",a_filename.c_str());
#endif

	m_frame=a_frame;
	m_spFilter=a_spFilter;

	pugi::xml_document document;
	pugi::xml_parse_result result=document.load_file(a_filename.c_str());
	if(!result)
	{
		return result;
	}

	startJoints();

#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load result \"%s\"\n",result.description());
#endif

	//* TODO check type and version
	const pugi::xml_node root=document.first_child();

	const pugi::xml_node collection=root.first_child();

	for(pugi::xml_node child=collection.first_child();child;
			child=child.next_sibling())
	{
		const String name=interpretString(child,"name");

		const String parentName=interpretString(child,"parent");

		const SpatialTransform refWorld=
				interpretTransform(child,"worldreference");
		const SpatialVector relativeTranslation=
				interpretVector(child,"translation");
		const SpatialTransform relativeRotation=
				interpretEuler(child,"rotation","rotationorder");
		const SpatialVector color=interpretVector(child,"color");

		SpatialTransform relativeMove;
		setIdentity(relativeMove);
		translate(relativeMove,relativeTranslation);

		const SpatialTransform xformRelative=relativeRotation*relativeMove;

#if FE_SAX_DEBUG
		const SpatialVector degrees=interpretVector(child,"rotation");
		const SpatialVector scale=interpretVector(child,"scale");

		feLog("\n%s \"%s\" parent \"%s\"\n",
				child.name(),name.c_str(),parentName.c_str());
		feLog("  T %s R %s\n  S %s\n",
				c_print(relativeTranslation),c_print(degrees),c_print(scale));
		feLog("  rot\n%s\n",c_print(relativeRotation));
		feLog("  ref\n%s\n",c_print(refWorld));
#endif

		addJointTweaked(name,parentName,xformRelative,refWorld,color);

	}

	return result;
}

SpatialTransform SurfaceAccessibleXml::interpretEuler(
	pugi::xml_node& a_rNode,String a_attrName,String a_attrOrder)
{
	const SpatialVector degrees=interpretVector(a_rNode,a_attrName);
	const String order=interpretString(a_rNode,a_attrOrder);

	const SpatialEuler euler(
			degrees[0]*degToRad,degrees[1]*degToRad,degrees[2]*degToRad);

	if(order=="ZYX")
	{
		return euler;
	}

	//* TODO move to Euler<>

	SpatialTransform xform;
	setIdentity(xform);

	if(order=="ZXY")
	{
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="XZY")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="YZX")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
	}
	else if(order=="XYZ")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
	}
	else if(order=="YXZ")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
	}

	return xform;
}

String SurfaceAccessibleXml::interpretString(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	return xmlAttr.value();
}

SpatialVector SurfaceAccessibleXml::interpretVector(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float x=0.0;
	float y=0.0;
	float z=0.0;
	sscanf(xmlAttr.value(),"%f,%f,%f",&x,&y,&z);

	return SpatialVector(x,y,z);
}

SpatialTransform SurfaceAccessibleXml::interpretTransform(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float v[16];
	sscanf(xmlAttr.value(),
			"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
			&v[0],&v[1],&v[2],&v[3],
			&v[4],&v[5],&v[6],&v[7],
			&v[8],&v[9],&v[10],&v[11],
			&v[12],&v[13],&v[14],&v[15]);

	SpatialTransform matrix;
	set(matrix,v);

	return matrix;
}

} // namespace
